# Copyright 2019-present Kensho Technologies, LLC.
from collections import namedtuple


RESERVED_PARAMETER_PREFIX = "_paged_"

# ParameterizedPaginationQueries namedtuple, describing two query ASTs that have PaginationFilters
# describing filters with which the query result size can be controlled. Note that these filters are
# returned parameterized i.e. values for the filters' parameters have yet to be generated.
# Additionally, a dict containing user-defined parameters is stored. Since this function may modify
# the user parameters to ensure better pagination, the user_parameters dict may differ from the
# original query's parameters that were provided to the paginator.
ParameterizedPaginationQueries = namedtuple(
    "ParameterizedPaginationQueries",
    (
        "next_page_query",  # Document, AST of query that will return the next page of
        # results when combined with pagination parameters.
        "remainder_query",  # Document, AST of query that will return the remainder of
        # results when combined with pagination parameters.
        "pagination_filters",  # List[PaginationFilter], filters usable for pagination.
        "user_parameters",  # dict, parameters that the user has defined for other filters.
    ),
)

# PaginationFilter namedtuples document filters usable for pagination purposes within the larger
# context of a ParameterizedPaginationQueries namedtuple. These filters may either be added by the
# query parameterizer, or filters that the user has added whose parameter values may be modified for
# generating paginated queries.
PaginationFilter = namedtuple(
    "PaginationFilter",
    (
        "vertex_class",  # str, vertex class to which the property field belongs to.
        "property_field",  # str, name of the property field filtering is done over.
        "next_page_query_filter",  # Directive, filter directive with '<' operator usable
        # for pagination in the page query.
        "remainder_query_filter",  # Directive, filter directive with '>=' operator usable
        # for pagination in the remainder query.
        "related_filters",  # List[Directive], filter directives that share the same
        # vertex and property field as the next_page_query_filter,
        # and are used to generate more accurate pages.
    ),
)


def generate_parameterized_queries(schema_info, query_ast, parameters):
    """Generate two parameterized queries that can be used to paginate over a given query.

    In order to paginate arbitrary GraphQL queries, additional filters may need to be added to be
    able to limit the number of results in the original query. This function creates two new queries
    with additional filters stored as PaginationFilters with which the query result size can be
    controlled.

    Args:
        schema_info: QueryPlanningSchemaInfo
        query_ast: Document, query that is being paginated.
        parameters: dict, list of parameters for the given query.

    Returns:
        ParameterizedPaginationQueries namedtuple
    """
    raise NotImplementedError()
